"""Module realise runtime checks for Luna services, databases, etc"""

import ast
import asyncio
from pathlib import Path
from typing import Awaitable, List, NamedTuple, Tuple

from log import requestIdCtx
from luna3.common.base_api import BaseHTTPAPI
from luna3.common.exceptions import LunaApiException

from ..constants.healthcheck import CHECK_TIMEOUT
from ..constants.paths import PROJECT_ROOT
from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from .functions import workingDirectory
from .log import Logger

logger = Logger("healthcheck")  # pylint: disable-msg=C0103


class Health(NamedTuple):
    """Health status structure"""

    name: str
    status: bool
    error: ErrorInfo

    def asDict(self):
        """Prepares dict representation."""
        return {"component": self.name, "error": self.error.asDict(), "status": int(self.status)}


async def inspectAlembicRevision(adaptor: "DBAdaptor") -> Tuple[str, str]:
    """
    Inspect alembic revisions. Get a real revision version from the database and an expected revision version
    from alembic versions files.
    Args:
        adaptor: database initialized adaptor
    Returns:
        tuple(<real migration version>, <expected migration version>)
    """

    from alembic.config import Config  # pylint: disable-msg=C0415
    from alembic.script import ScriptDirectory  # pylint: disable-msg=C0415
    from sqlalchemy import text  # pylint: disable-msg=C0415

    def getMigrationTableName(pathToAlembicEnv):
        alembicEnvModule = ast.parse(Path(pathToAlembicEnv).read_text(encoding="utf-8"))
        for expression in alembicEnvModule.body:
            if not isinstance(expression, ast.Assign):
                continue
            if len(expression.targets) != 1:
                continue
            if not isinstance(expression.targets[0], ast.Name):
                continue
            if expression.targets[0].id == "VERSION_TABLE_NAME":
                if isinstance(expression.value, ast.Constant):
                    # python >=3.8
                    migrationTableName = expression.value.value
                elif isinstance(expression.value, ast.Str):
                    # python 3.7
                    migrationTableName = expression.value.s
                else:
                    raise RuntimeError(f"bad type of expression value {expression.value}")
                break
        else:
            raise RuntimeError("Failed to find a table migration name")
        return migrationTableName

    with workingDirectory(PROJECT_ROOT):
        alembicCfg = Config("alembic.ini")
        script = ScriptDirectory.from_config(alembicCfg)
        expectedMigration = script.get_revisions(script.get_heads())[0].revision
        migrationTable = getMigrationTableName(script.env_py_location)

    async with adaptor.connection(logger) as connection:
        realMigration = await connection.scalar(text(f'select * from "{migrationTable}"'))
    return realMigration, expectedMigration


async def checkSqlMigration(dbAdaptor: "DBAdaptor"):
    """
    Check database revision version.

    Args:
        dbAdaptor: DBAdaptor instance

    Raises:
        HealthcheckError if versions dont match.
    """
    currentRevision, expectedRevision = await inspectAlembicRevision(adaptor=dbAdaptor)

    if currentRevision != expectedRevision:
        errorMsg = (
            f"Actual and expected db revisions are not equal (actual '{currentRevision}', expected "
            f"'{expectedRevision}'); please check that you have executed the database migration"
        )
        raise VLException(Error.HealthCheckError.format(errorMsg), 500, isCriticalError=False)


async def checkSql(dbAdaptor: "DBAdaptor"):
    """Check database availability."""
    from sqlalchemy import select as saSelect  # pylint: disable-msg=C0415

    async with dbAdaptor.connection(logger) as connection:
        await connection.execute(saSelect([1]))


async def checkRedis(client: "Redis"):
    """
    Redis health check function. Pings redis.

    Args:
        client: Either Redis or ConnectionsPool instance
    """
    try:
        await asyncio.wait_for(client.ping(), timeout=CHECK_TIMEOUT)  # Remove timeout after upgrade to aioredis 2.x
    except asyncio.TimeoutError as exc:
        raise VLException(Error.DataBaseConnectionTimeoutError, 500, False) from exc


async def checkService(client: BaseHTTPAPI, includeLunaServices: bool = False):
    """
    Luna service check function.

    Args:
        client: configured service client (i.e. luna3Session.getClient(rid).lunaFaces)
        includeLunaServices: whether to perform checks for dependencies of the service
    """
    requestId = requestIdCtx.get()
    checker = client.checkHealth(
        includeLunaServices=includeLunaServices,
        lunaRequestId=requestId,
        totalTimeout=CHECK_TIMEOUT,
        asyncRequest=True,
        raiseError=True,
    )
    await checker


async def performCheck(name: str, check: Awaitable) -> List[Health]:
    """
    Process given check.

    Check is considered successful if no exceptions were raised.
    Check is responsible for handling timeouts.

    Args:
        name: str, identifier of the check
        check: coroutine to run

    Returns:
        Health struct
    """
    try:
        await check
    except VLException as exc:
        if exc.isCriticalError:
            logger.exception()
        error = Error.HealthCheckError.format(exc.error.detail)
    except LunaApiException as exc:
        if exc.statusCode == 502:
            results = []
            for res in exc.json["errors"]:
                error = ErrorInfo.fromDict(res["error"])
                results.append(Health(f"{name}.{res['component']}", error=error, status=error == Error.Success))
            return results

        if exc.statusCode == 504:
            # timeout, doesn't care wich one (should be TOTAL)
            logger.exception()
            error = Error.HealthCheckError.format("Timed out")
        else:
            logMsg = exc.json["detail"]
            error = Error.HealthCheckError.format(logMsg)
    else:
        error = Error.Success

    return [Health(name, error=error, status=error is Error.Success)]
